import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestVibratingPendulum(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(8)
    @timeout_decorator.timeout(5.)
    def test_q1_pendulum_controller(self):
        """Test pendulum controller"""
        # note: all prints here go to the output item in the json file

        # the response must be monotonically increasing
        theta_dot = self.notebook_locals['logger'].data()[-1, :]
        self.assertTrue(
            all(i <= j + 1e-4 for i, j in zip(theta_dot, theta_dot[1:])),
            msg='theta_dot(t) is decreasing.',
        )

        # final theta dot in [.99, 1.001]
        self.assertLessEqual(
            theta_dot[-1],
            1.001,
            msg='theta_dot(t=10) = {} > 1.001'.format(theta_dot[-1]),
        )
        self.assertGreaterEqual(
            theta_dot[-1],
            .99,
            msg='theta_dot(t=10) = {} < 0.99'.format(theta_dot[-1]),
        )



import numpy as np
import matplotlib.pyplot as plt

from pydrake.systems.pyplot_visualizer import PyPlotVisualizer
from pydrake.systems.framework import Context, PortDataType

_MIT_RED = "#A31F34"


class Brick(object):
    WIDTH = 3.
    HEIGHT = 1.

    def __init__(self):
        self.patch = plt.Rectangle((0.0, 0.0),
                                   self.WIDTH,
                                   self.HEIGHT,
                                   fc=_MIT_RED,
                                   ec="k")
        self.set_state(0.)

    def set_state(self, x):
        self.patch.set_x(x - self.WIDTH / 2)  # Center at x

    @classmethod
    def add_to_axes(cls, ax):
        brick = cls()
        ax.add_patch(brick.patch)
        return brick


class DoubleIntegratorVisualizer(PyPlotVisualizer):

    # Limits of view port
    XLIM = (-15., 15.)
    YLIM = (-6., 6.)
    TICK_DIMS = (0.2, 0.8)

    def __init__(self, ax=None, show=None):
        PyPlotVisualizer.__init__(self, ax=ax, show=show)
        self.DeclareInputPort("state", PortDataType.kVectorValued, 2)

        self.ax.set_xlim(*self.XLIM)
        self.ax.set_ylim(*self.YLIM)
        self.ax.set_aspect("auto")

        self._make_background()
        self.brick = Brick.add_to_axes(self.ax)

    def _make_background(self):
        # x-axis
        self.ax.plot(self.XLIM, np.zeros_like(self.XLIM), "k")
        # tick mark centered at the origin
        tick_pos = -0.5 * np.asarray(self.TICK_DIMS)
        self.ax.add_patch(plt.Rectangle(tick_pos, *self.TICK_DIMS, fc="k"))

    def draw(self, context):
        try:
            x = self.EvalVectorInput(context, 0).get_value()[0]
            self.ax.set_title("t = {:.1f}".format(context.get_time()))
        except TypeError:
            x = context[0]
        self.brick.set_state(x)



import os


def FindResource(filename):
    return os.path.join(os.path.dirname(__file__), filename)


def Rgba2Hex(rgb):
    """
    Turn a list of R,G,B elements (any indexable list of >= 3 elements will
    work), where each element is specified on range [0., 1.], into the
    equivalent 24-bit value 0xRRGGBB.
    """
    val = 0
    for i in range(3):
        val += (256**(2 - i)) * int(255 * rgb[i])
    return val



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestCompassGaitLimitCycle(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    # starts by testing the constraints we already implemented
    # but gives zero points for each test

    def _test_time_intervals(self):
        """Time intervals are within bounds"""
        # retrieve variables from notebook
        T = self.notebook_locals['T']
        h_opt = self.notebook_locals['h_opt']
        h_min = self.notebook_locals['h_min']
        h_max = self.notebook_locals['h_max']

        # check one per time
        tol = 1e-4
        for t in range(T):
            self.assertTrue(h_opt[t] >= h_min - tol,
                            msg=f'Time interval lower bound violated at time '
                            + f'step {t}.')
            self.assertTrue(h_opt[t] <= h_max + tol,
                            msg=f'Time interval upper bound violated at time '
                            + f'step {t}.')

    def _test_dynamic_feasibility(self):
        """Implicit Euler equations are verified"""
        # retrieve variables from notebook
        T = self.notebook_locals['T']
        h = self.notebook_locals['h_opt']
        q = self.notebook_locals['q_opt']
        qd = self.notebook_locals['qd_opt']
        qdd = self.notebook_locals['qdd_opt']
        f = self.notebook_locals['f_opt']
        manipulator_equations = self.notebook_locals['manipulator_equations']

        # link configurations, velocities, accelerations using implicit Euler
        for t in range(T):
            np.testing.assert_allclose(
                q[t + 1],
                q[t] + h[t] * qd[t + 1],
                atol=1e-5,
                rtol=1e-5,
                err_msg='Configurations q and velocities qd do not verify '
                + f'the Implicit Euler rule at time step {t}.')
            np.testing.assert_allclose(
                qd[t + 1],
                qd[t] + h[t] * qdd[t],
                atol=1e-5,
                rtol=1e-5,
                err_msg='Velocities qd and accelerations qdd do not verify '
                + f'the Implicit Euler rule at time step {t}.')
            vars = np.concatenate((q[t + 1], qd[t + 1], qdd[t], f[t]))
            manipulator_equations_violation = manipulator_equations(vars)
            np.testing.assert_allclose(
                manipulator_equations_violation,
                np.zeros(q.shape[1]),
                atol=1e-5,
                rtol=1e-5,
                err_msg='Manipulator equations are not verified at time '
                + f'step {t}.')

    def _test_impulsive_collision(self):
        """Impulsive collision is verified at the heel strike"""
        # retrieve variables from notebook
        q = self.notebook_locals['q_opt']
        qd = self.notebook_locals['qd_opt']
        qd_post = self.notebook_locals['qd_post_opt']
        imp = self.notebook_locals['imp_opt']
        reset_velocity_heelstrike = self.notebook_locals[
            'reset_velocity_heelstrike']

        # link pre and post impact velocities
        vars = np.concatenate((q[-1], qd[-1], qd_post, imp))
        reset_velocity_heelstrike_violation = reset_velocity_heelstrike(vars)
        np.testing.assert_allclose(
            reset_velocity_heelstrike_violation,
            np.zeros(q.shape[1] + imp.size),
            atol=1e-5,
            rtol=1e-5,
            err_msg='Impulsive collision equations are not verified.')

    def _test_periodicity(self):
        """Periodicity of the walking gait"""
        # retrieve variables from notebook
        q = self.notebook_locals['q_opt']
        qd = self.notebook_locals['qd_opt']
        qd_post = self.notebook_locals['qd_post_opt']

        # periodicity of the configuration
        np.testing.assert_allclose(
            q[0],
            -q[-1],
            atol=1e-5,
            rtol=1e-5,
            err_msg='Configuration vector q does not verify periodicity '
            + 'conditions.')

        # periodicity of the  velocities
        np.testing.assert_allclose(
            qd[0],
            np.array([0, 0, qd_post[2] + qd_post[3], -qd_post[3]]),
            atol=1e-5,
            rtol=1e-5,
            err_msg='Velocity vector qd does not verify periodicity '
            + 'conditions.')

    # tests constraints implemented by the students
    # does not use the same expressions that must be used in the constraint
    # e.g. checks that a linear constraint is verified by evaluating a nonlinear
    # function

    @weight(3)
    @timeout_decorator.timeout(3.)
    def test_stance_foot_on_ground(self):
        """Stance foot on the ground for all times"""
        # run the tests for the constraints already in the notebook
        self._test_time_intervals()
        self._test_dynamic_feasibility()
        self._test_impulsive_collision()
        self._test_periodicity()

        # retrieve variables from notebook
        T = self.notebook_locals['T']
        q = self.notebook_locals['q_opt']
        for t in range(T + 1):
            self.assertAlmostEqual(
                np.linalg.norm(q[t, :2]),
                0,
                msg=f'Stance foot is not on the ground at time step {t}.')

    @weight(2)
    @timeout_decorator.timeout(3.)
    def test_swing_foot_on_ground(self):
        """Swing foot on the ground at time zero"""
        # run the tests for the constraints already in the notebook
        self._test_time_intervals()
        self._test_dynamic_feasibility()
        self._test_impulsive_collision()
        self._test_periodicity()

        # retrieve variables from notebook
        q = self.notebook_locals['q_opt']
        swing_foot_height = self.notebook_locals['swing_foot_height']
        self.assertAlmostEqual(
            swing_foot_height(q[0])[0],
            0,
            msg='Swing foot is not on the ground at time zero.')

    @weight(3)
    @timeout_decorator.timeout(3.)
    def test_no_penetration(self):
        """No penetration of the swing foot in the ground for all times"""
        # run the tests for the constraints already in the notebook
        self._test_time_intervals()
        self._test_dynamic_feasibility()
        self._test_impulsive_collision()
        self._test_periodicity()

        # retrieve variables from notebook
        T = self.notebook_locals['T']
        q = self.notebook_locals['q_opt']
        swing_foot_height = self.notebook_locals['swing_foot_height']
        tol = 1e-4
        for t in range(T + 1):
            self.assertGreater(
                swing_foot_height(q[t])[0],
                -tol,
                msg=f'Swing foot in penetration with ground at time step {t}.')

    @weight(3)
    @timeout_decorator.timeout(3.)
    def test_stance_foot_friction(self):
        """Stance-foot contact force in friction cone for all times"""
        # run the tests for the constraints already in the notebook
        self._test_time_intervals()
        self._test_dynamic_feasibility()
        self._test_impulsive_collision()
        self._test_periodicity()

        # retrieve variables from notebook
        T = self.notebook_locals['T']
        friction = self.notebook_locals['friction']
        f = self.notebook_locals['f_opt']
        tol = 1e-4
        for t in range(T):
            self.assertGreater(
                f[t, 1],
                -tol,
                msg='Stance-foot normal contact force is negative at time '
                + f'step {t}.')
            self.assertLess(
                np.abs(f[t, 0]),
                friction * f[t, 1] + tol,
                msg='Stance-foot contact force outside friction cone at time '
                + f'step {t}.')

    @weight(2)
    @timeout_decorator.timeout(3.)
    def test_swing_foot_friction(self):
        """Swing-foot impulse in friction cone"""
        # run the tests for the constraints already in the notebook
        self._test_time_intervals()
        self._test_dynamic_feasibility()
        self._test_impulsive_collision()
        self._test_periodicity()

        # retrieve variables from notebook
        friction = self.notebook_locals['friction']
        imp = self.notebook_locals['imp_opt']
        tol = 1e-4
        self.assertGreater(
            imp[1],
            -tol,
            msg='Normal heel-strike impulse is negative at time step {t}.')
        self.assertLess(np.abs(imp[0]),
                        friction * imp[1] + tol,
                        msg='Heel-strike impulse outside friction cone')



import numpy as np
import pydrake.all
from matplotlib.colors import Colormap

import meshcat.geometry as g
import meshcat.transformations as tf

from pydrake.all import RigidTransform


# TODO: Remove this once https://github.com/rdeits/meshcat-python/pull/91 makes
# its way through to drake.
class TriangularMeshGeometry(g.Geometry):
    """
    A mesh consisting of an arbitrary collection of triangular faces. To
    construct one, you need to pass in a collection of vertices as an Nx3 array
    and a collection of faces as an Mx3 array. Each element of `faces` should
    be a collection of 3 indices into the `vertices` array.

    For example, to create a square made out of two adjacent triangles, we
    could do:

    vertices = np.array([
        [0, 0, 0],  # the first vertex is at [0, 0, 0]
        [1, 0, 0],
        [1, 0, 1],
        [0, 0, 1]
    ])
    faces = np.array([
        [0, 1, 2],  # The first face consists of vertices 0, 1, and 2
        [3, 0, 2]
    ])

    mesh = TriangularMeshGeometry(vertices, faces)

    To set the color of the mesh by vertex, pass an Nx3 array containing the
    RGB values (in range [0,1]) of the vertices to the optional `color`
    argument, and set `vertexColors=True` in the Material.
    """
    __slots__ = ["vertices", "faces"]

    def __init__(self, vertices, faces, color=None):
        super(TriangularMeshGeometry, self).__init__()

        vertices = np.asarray(vertices, dtype=np.float32)
        faces = np.asarray(faces, dtype=np.uint32)
        assert vertices.shape[1] == 3, "`vertices` must be an Nx3 array"
        assert faces.shape[1] == 3, "`faces` must be an Mx3 array"
        self.vertices = vertices
        self.faces = faces
        if color is not None:
            color = np.asarray(color, dtype=np.float32)
            assert np.array_equal(vertices.shape, color.shape),\
                "`color` must be the same shape as vertices"
        self.color = color

    def lower(self, object_data):
        attrs = {u"position": g.pack_numpy_array(self.vertices.T)}
        if self.color is not None:
            attrs[u"color"] = g.pack_numpy_array(self.color.T)
        return {
            u"uuid": self.uuid,
            u"type": u"BufferGeometry",
            u"data": {
                u"attributes": attrs,
                u"index": g.pack_numpy_array(self.faces.T)
            }
        }


def plot_surface(meshcat, X, Y, Z, color=0xdd9999, wireframe=False):
    (rows, cols) = Z.shape

    vertices = np.empty((rows * cols, 3), dtype=np.float32)
    vertices[:, 0] = X.reshape((-1))
    vertices[:, 1] = Y.reshape((-1))
    vertices[:, 2] = Z.reshape((-1))

    # Vectorized faces code from https://stackoverflow.com/questions/44934631/making-grid-triangular-mesh-quickly-with-numpy  # noqa
    faces = np.empty((rows - 1, cols - 1, 2, 3), dtype=np.uint32)
    r = np.arange(rows * cols).reshape(rows, cols)
    faces[:, :, 0, 0] = r[:-1, :-1]
    faces[:, :, 1, 0] = r[:-1, 1:]
    faces[:, :, 0, 1] = r[:-1, 1:]
    faces[:, :, 1, 1] = r[1:, 1:]
    faces[:, :, :, 2] = r[1:, :-1, None]
    faces.shape = (-1, 3)

    if isinstance(color, Colormap):
        z = vertices[:, 2]
        rgba = color((vertices[:, 2] - np.min(z)) / np.ptp(z))
        color = rgba[:, :3]

    if isinstance(color, int):
        meshcat.set_object(
            g.TriangularMeshGeometry(vertices, faces),
            g.MeshLambertMaterial(color=color, wireframe=wireframe))
    else:
        meshcat.set_object(
            TriangularMeshGeometry(vertices, faces, color),
            g.MeshLambertMaterial(vertexColors=True, wireframe=wireframe))


def plot_mathematical_program(meshcat,
                              prog,
                              X,
                              Y,
                              result=None,
                              point_size=0.05):
    assert prog.num_vars() == 2
    assert X.size == Y.size

    N = X.size
    values = np.vstack((X.reshape(-1), Y.reshape(-1)))
    costs = prog.GetAllCosts()

    # Vectorized multiply for the quadratic form.
    # Z = (D*np.matmul(Q,D)).sum(0).reshape(nx, ny)

    if costs:
        Z = prog.EvalBindingVectorized(costs[0], values)
        for b in costs[1:]:
            Z = Z + prog.EvalBindingVectorized(b, values)

    cv = meshcat["constraints"]
    for binding in prog.GetAllConstraints():
        if isinstance(
                binding.evaluator(),
                pydrake.solvers.mathematicalprogram.BoundingBoxConstraint):
            c = binding.evaluator()
            var_indices = [
                int(prog.decision_variable_index()[v.get_id()])
                for v in binding.variables()
            ]
            satisfied = np.array(
                c.CheckSatisfiedVectorized(values[var_indices, :],
                                           0.001)).reshape(1, -1)
            if costs:
                Z[~satisfied] = np.nan

            v = cv[type(c).__name__]
            Zc = np.zeros(Z.shape)
            Zc[satisfied] = np.nan
            plot_surface(v,
                         X,
                         Y,
                         Zc.reshape((X.shape[1], X.shape[0])),
                         color=0xff3333,
                         wireframe=True)
        else:
            Zc = prog.EvalBindingVectorized(binding, values)
            evaluator = binding.evaluator()
            low = evaluator.lower_bound()
            up = evaluator.upper_bound()
            cvb = cv[type(evaluator).__name__]
            for index in range(Zc.shape[0]):
                color = np.repeat([[0.3, 0.3, 1.0]], N, axis=0)
                infeasible = np.logical_or(Zc[index, :] < low[index],
                                           Zc[index, :] > up[index])
                color[infeasible, 0] = 1.0
                color[infeasible, 2] = 0.3
                plot_surface(cvb[str(index)],
                             X,
                             Y,
                             Zc[index, :].reshape(X.shape[1], X.shape[0]),
                             color=color,
                             wireframe=True)

    if costs:
        plot_surface(meshcat["objective"],
                     X,
                     Y,
                     Z.reshape(X.shape[1], X.shape[0]),
                     color=0x77cc77,
                     wireframe=True)

    if result:
        v = meshcat["solution"]
        v.set_object(g.Sphere(point_size),
                     g.MeshLambertMaterial(color=0x55ff55))
        x_solution = result.get_x_val()
        v.set_transform(
            tf.translation_matrix(
                [x_solution[0], x_solution[1],
                 result.get_optimal_cost()]))


# pulled from inside MeshcatVisualizer.
def set_planar_viewpoint(vis,
                         camera_position=[0, -1, 0],
                         camera_focus=[0, 0, 0],
                         xmin=-1,
                         xmax=1,
                         ymin=-1,
                         ymax=1):
    # TODO(russt): Figure out the proper set of camera transformations to
    # implement camera_focus.
    if np.any(camera_focus):
        warnings.warn("Non-zero camera_focus is not supported yet")

    # Set up orthographic camera.
    camera = g.OrthographicCamera(left=xmin,
                                  right=xmax,
                                  top=ymax,
                                  bottom=ymin,
                                  near=-1000,
                                  far=1000)
    vis['/Cameras/default/rotated'].set_object(camera)
    vis['/Cameras/default'].set_transform(
        RigidTransform(camera_position).GetAsMatrix4())

    # Lock the orbit controls.
    vis['/Cameras/default/rotated/<object>'].set_property("position", [0, 0, 0])

    # Turn off background, axes, and grid.
    vis['/Background'].set_property("visible", False)
    vis['/Grid'].set_property("visible", False)
    vis['/Axes'].set_property("visible", False)


def draw_points(meshcat, points, color, **kwargs):
    """Helper for sending a 3xN points of a single color to MeshCat"""
    points = np.asarray(points)
    assert points.shape[0] == 3
    if points.size == 3:
        points.shape = (3, 1)
    colors = np.tile(np.asarray(color).reshape(3, 1), (1, points.shape[1]))
    meshcat.set_object(g.PointCloud(points, colors, **kwargs))



from copy import copy
import numpy as np
import matplotlib.pyplot as plt

from pydrake.all import System


def plot_2d_phase_portrait(f, x1lim=(-1, 1), x2lim=(-1, 1), n=100j, **kwargs):
    """
    Plots the phase portrait for a 2D dynamical system.

    Parameters
    ----------
    f : function
        Callable vectorized function which implements the rhs of the
        state-space dynamics xdot = f(x), with x 2D array.
    x1lim : tuple
        Minimum and maximum values (floats) for the horizontal axis of the
        plot.
    x2lim : tuple
        Minimum and maximum values (floats) for the vertical axis of the
        plot.
    n : complex
        Purely imaginary number with imaginary part equal to the number of knot
        points on each axis of the plot.
    """
    # grid state space, careful here x2 before x1
    X1, X2 = np.mgrid[x1lim[0]:x1lim[1]:n, x2lim[0]:x2lim[1]:n]
    if isinstance(f, System):
        context = f.CreateDefaultContext()
        X1d = copy(X1)
        X2d = copy(X2)
        for i in range(X1.shape[0]):
            for j in range(X1.shape[1]):
                context.SetContinuousState([X1[i, j], X2[i, j]])
                deriv = f.EvalTimeDerivatives(context).CopyToVector()
                X1d[i, j] = deriv[0]
                X2d[i, j] = deriv[1]
    else:
        X1d, X2d = f([X1, X2])

    # color the streamlines according to the magnitude of xdot
    color = np.sqrt(X1d**2 + X2d**2)

    # phase portrait (annoying input format of streamplot)
    strm = plt.streamplot(X1.T[0], X2[0], X1d.T, X2d.T, color=color.T, **kwargs)

    # colorbar on the right that measures the magnitude of xdot
    plt.gcf().colorbar(strm.lines, label=r'$|\dot\mathbf{x}|$')

    # misc plot settings
    plt.xlabel(r'$x_1$')
    plt.ylabel(r'$x_2$')
    plt.xlim(x1lim)
    plt.ylim(x2lim)
    plt.gca().set_aspect('equal')



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestFootstepPlanning(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        """
        Since the planner raises an error if the problem is infeasible, it is
        better to solve the optimizations inside each method rather than here in
        the __init__ method. This way if the plan is infeasible, students get
        zero points instead of an error from the test file. Note also that each
        time a test (method of this class) fails a new class is instantiated,
        hence we must re-solve the plans for each test.
        """
        super().__init__(test_name)

        # get locals
        self.notebook_locals = notebook_locals
        Terrain = self.notebook_locals['Terrain']
        self.footstep_planner = self.notebook_locals['footstep_planner']

        # empty solution of the benchmark problems
        self.terrains = {'A': Terrain([1, 1, 1, 1]), 'B': Terrain([1, 1, 1, 0])}
        self.n_steps = 8
        self.step_span = .8
        self.plans = {'A': {}, 'B': {}}

    def _solve_benchmarks(self):
        """Solves the benchmark problems."""
        for plan in ['A', 'B']:
            self.plans[plan]['vars'], self.plans[plan][
                'cost'] = self.footstep_planner(self.terrains[plan],
                                                self.n_steps, self.step_span)

    @weight(4)
    @timeout_decorator.timeout(30.)
    def test_relative_position_limits(self):
        """Tests the constraints on the relative position of the feet"""
        self._solve_benchmarks()

        # center to corner vector for the square with side step_span
        # (plus tolerance for the checks)
        tol = 1e-2
        c2c = np.ones(2) * self.step_span / 2 + tol

        # check constraint for both terrains
        for plan in ['A', 'B']:
            position_left, position_right = self.plans[plan]['vars'][:2]
            relative_position = position_left - position_right
            for t in range(self.n_steps + 1):
                msg = f'Error at foot step {t} when solving the problem ' + \
                    f'for terrain_{plan}: one foot is not contained in a ' + \
                    f'square of side {self.step_span} centered at the ' + \
                    'other foot.'
                np.testing.assert_array_less(relative_position[t], c2c, msg)
                np.testing.assert_array_less(-relative_position[t], c2c, msg)

    @weight(3)
    @timeout_decorator.timeout(30.)
    def test_one_stone_per_foot(self):
        """Tests that a foot is assigned to a single stepping stone"""
        self._solve_benchmarks()

        # check constraint for both terrains
        for plan in ['A', 'B']:
            stone_left, stone_right = self.plans[plan]['vars'][2:4]
            for t in range(self.n_steps + 1):
                msg = f'Error at foot step {t} when solving the problem ' + \
                    f'for terrain_{plan}: '
                self.assertAlmostEqual(
                    sum(stone_left[t]),
                    1,
                    places=4,
                    msg=msg + f'stone_left[{t}] do not sum up to one.')
                self.assertAlmostEqual(
                    sum(stone_right[t]),
                    1,
                    places=4,
                    msg=msg + f'stone_right[{t}] do not sum up to one.')

    @weight(5)
    @timeout_decorator.timeout(30.)
    def test_foot_in_stepping_stone(self):
        """Tests that binaries imply correctly the positioning of the feet"""
        self._solve_benchmarks()
        tol = 1e-2

        # check constraint for both terrains
        for plan in ['A', 'B']:
            position_left, position_right, \
                stone_left, stone_right = self.plans[plan]['vars'][:4]
            for t in range(self.n_steps + 1):

                # left foot
                i = np.where(np.isclose(stone_left[t], 1, atol=tol))[0][0]
                stone = self.terrains[plan].stepping_stones[i]
                np.testing.assert_array_less(
                    stone.A.dot(position_left[t]), stone.b + tol,
                    f'Solving the problem for terrain_{plan}, the binary '
                    + f'variable stone_left[{t}, {i}] is one but the left foot '
                    + f'is not in stone {i} at time {t}.')

                # right foot
                i = np.where(np.isclose(stone_right[t], 1, atol=tol))[0][0]
                stone = self.terrains[plan].stepping_stones[i]
                np.testing.assert_array_less(
                    stone.A.dot(position_right[t]), stone.b + tol,
                    f'Solving the problem for terrain_{plan}, the binary '
                    + f'variable stone_right[{t}, {i}] is one but the right '
                    + f'foot is not in stone {i} at time {t}.')

    @weight(4)
    @timeout_decorator.timeout(30.)
    def test_minimize_step_length(self):
        """
        Tests that the positioning of the feet minimized the sum of the
        square of the two norm of the step lengths.
        """
        self._solve_benchmarks()
        tol = 1e-2

        # check constraint for both terrains
        target_cost = {'A': 4.1592510194, 'B': 5.6478552756}
        for plan in ['A', 'B']:

            # same convergence check as in
            # https://github.com/RobotLocomotion/drake/blob/88c93118df507777eb3f99628d1aa7c808f81f49/solvers/branch_and_bound.cc#L711
            # with double the tolerances from
            # https://github.com/RobotLocomotion/drake/blob/4ee674e7931527df838bd33e79cf2f4dad57bd20/solvers/branch_and_bound.h#L674
            atol = 2e-2
            rtol = 2e-2
            gap = self.plans[plan]['cost'] - target_cost[plan]
            atol_convergence = gap <= atol
            rtol_convergence = gap / target_cost[plan] <= rtol
            self.assertTrue(
                atol_convergence or rtol_convergence,
                f'Target cost of {target_cost[plan]} is not achieved for '
                + f'terrain_{plan}, with absolute tolerance {atol} and '
                + f'relative tolerance {rtol}.')



import unittest
from gradescope_utils.autograder_utils.json_test_runner import JSONTestRunner
import json
from runpy import run_path
import nbformat
from nbconvert import PythonExporter
import os


class Grader:

    def __init__(self):
        pass

    @staticmethod
    def grade_from_notebooks(test_cases_list, notebook_ipynb_list,
                             results_json):
        """
        Running notebooks in notebook_ipynb_list and evaluating
        them on test_cases_list. Result is written into results_json
        """
        try:
            notebook_locals_list = []
            for notebook_ipynb in notebook_ipynb_list:
                notebook_locals_list.append(
                    Grader.locals_from_notebook(notebook_ipynb))
        except Exception as e:
            Grader.global_fail_with_error_message(
                "Exception when running file: " + notebook_ipynb + ', '
                + str(e), results_json)
            raise

        # Grade notebook_locals_list on test_cases_list
        Grader.grade_output(test_cases_list, notebook_locals_list, results_json)

    @staticmethod
    def grade_output(test_case_list, notebook_locals_list, results_json):
        """Grading the notebook_locals with the provided test_cases"""
        # setup test suite for gradescope
        suite = unittest.TestSuite()
        for test_case, notebook_locals in zip(test_case_list,
                                              notebook_locals_list):
            test_case_names = unittest.defaultTestLoader.getTestCaseNames(
                test_case)
            for test_case_name in test_case_names:
                suite.addTest(test_case(test_case_name, notebook_locals))

        # run all the tests in the suite
        with open(results_json, 'w') as fh:
            JSONTestRunner(stream=fh).run(suite)

    @staticmethod
    def locals_from_notebook(notebook_ipynb):
        """Read, run, return locals of notebook"""
        banned_commands = ['HTML']

        ipynb = json.load(open(notebook_ipynb))

        for cell in ipynb['cells']:
            # erase test cells, this is optional and useful for debugging
            # to avoid recursions when developing
            if any('## TEST ##' in line for line in cell['source']):
                cell['source'] = []
            # filter out all the lines with banned commands
            if banned_commands is not None:
                cell['source'] = [
                    line for line in cell['source']
                    if not any(command in line for command in banned_commands)
                ]

        exporter = PythonExporter()
        source, meta = exporter.from_notebook_node(
            nbformat.reads(json.dumps(ipynb), nbformat.NO_CONVERT))
        with open('./cleaned_notebook.py', 'w') as fh:
            fh.write(source)

        notebook_locals = run_path('cleaned_notebook.py')
        os.system('rm cleaned_notebook.py')
        return notebook_locals

    @staticmethod
    def global_fail_with_error_message(msg, results_json):
        """Error message if no specific"""
        results = {"score": 0.0, "output": msg}

        with open(results_json, 'w') as f:
            f.write(
                json.dumps(results,
                           indent=4,
                           sort_keys=True,
                           separators=(',', ': '),
                           ensure_ascii=True))

    @staticmethod
    def print_test_results(results_json):
        """Printing the results.json"""
        # open the json file for reading
        with open(results_json, 'r') as fh:
            result = json.load(fh)

        # print total score
        max_score = sum(test['max_score'] for test in result['tests'])
        print('Total score is {}/{}.'.format(int(result['score']), max_score))

        # print partial scores
        for test in result['tests']:
            print('\nScore for {} is {}/{}.'.format(test['name'],
                                                    int(test['score']),
                                                    test['max_score']))

            # print error message if any
            if 'output' in test:
                print('- ' + test['output'])



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestIlqrDriving(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(1)
    @timeout_decorator.timeout(1.)
    def test1_discrete_dynamics(self):
        """Test discrete_dynamics"""
        # load locals
        f = self.notebook_locals['discrete_dynamics']

        f_target = [
            np.array(
                [0.14181286, 0.81062987, 0.54580749, 0.77731476, 1.02810156]),
            np.array([0.131663, 0.300996, 0.57034388, 0.71732411, 0.81033267]),
            np.array(
                [0.33233256, 0.91916814, 0.27416228, 0.45461388, 0.99126091]),
            np.array(
                [1.02770659, 0.27770476, 0.56066881, 0.96146963, 0.20821043]),
            np.array([0.71706372, 0.47773599, 0.22401885, 0.538506, 0.40897373])
        ]

        np.random.seed(7)
        n_rands = 5
        f_eval = []
        for i in range(n_rands):
            x = np.random.rand(5)
            u = np.random.rand(2)
            f_eval = f(x, u)
            np.testing.assert_array_almost_equal(
                f_eval,
                f_target[i],
                err_msg='The discrete dynamics are incorrect. '
                f'Expected output {f_target[i]} with x={x} and u={u}')

    @weight(1)
    @timeout_decorator.timeout(1.)
    def test2_rollout(self):
        """Test rollout"""
        # load locals
        f = self.notebook_locals['rollout']

        f_target_hash = [
            5548026545464297818, 5303601661710898095, -8752618797308043950,
            5303655996741254866, 1117140169305939251
        ]

        np.random.seed(7)
        n_rands = 5
        for i in range(n_rands):
            N = 5
            n_u = 2
            x0 = np.array([1, 0, 0, 1, 0])
            u_trj = np.random.randn(N - 1, n_u) * 0.4
            f_eval_hash = hash(
                tuple(np.ndarray.flatten(self.round(f(x0, u_trj), 4))))
            self.assertEqual(f_eval_hash, f_target_hash[i], "Incorrect rollout")

    @weight(1)
    @timeout_decorator.timeout(1.)
    def test3_cost_trj(self):
        """Test cost_trj"""
        # load locals
        f = self.notebook_locals['cost_trj']

        f_target = [
            29.946731429035616, 34.805336401404986, 27.626872505986434,
            33.25730851334137, 33.83897480138296
        ]

        np.random.seed(7)
        n_rands = 5
        f_eval = []
        for i in range(n_rands):
            N = 5
            n_u = 2
            n_x = 5
            x0 = np.array([1, 0, 0, 1, 0])
            u_trj = np.random.randn(N - 1, n_u) * 0.4
            x_trj = np.random.randn(N, n_x) * 0.4
            f_eval = f(x_trj, u_trj)
            np.testing.assert_almost_equal(
                f_eval,
                f_target[i],
                err_msg='The cost_trj function is incorrect. '
                f'Expected output {f_target[i]} with x_trj=\n{x_trj}'
                f' and u=\n{u_trj}')

    @weight(5)
    @timeout_decorator.timeout(1.)
    def test4_Q_terms(self):
        """Test Q_terms"""
        # load locals
        Q_terms = self.notebook_locals['Q_terms']
        derivs = self.notebook_locals['derivs']

        np.random.seed(7)
        x = np.random.rand(5)
        u = np.random.rand(2)

        l_x, l_u, l_xx, l_ux, l_uu, f_x, f_u = derivs.stage(x, u)
        V_x, V_xx = derivs.final(x + 0.1)
        Q_x, Q_u, Q_xx, Q_ux, Q_uu = Q_terms(l_x, l_u, l_xx, l_ux, l_uu, f_x,
                                             f_u, V_x, V_xx)

        Q_x_hash = hash(tuple(np.ndarray.flatten(self.round(Q_x, 4))))
        Q_u_hash = hash(tuple(np.ndarray.flatten(self.round(Q_u, 4))))
        Q_xx_hash = hash(tuple(np.ndarray.flatten(self.round(Q_xx, 4))))
        Q_ux_hash = hash(tuple(np.ndarray.flatten(self.round(Q_ux, 4))))
        Q_uu_hash = hash(tuple(np.ndarray.flatten(self.round(Q_uu, 4))))

        self.assertEqual(Q_x_hash, 7805697548178880088, "Incorrect Q_x")
        self.assertEqual(Q_u_hash, 980374240801014371, "Incorrect Q_u")
        self.assertEqual(Q_xx_hash, 7557069004417576783, "Incorrect Q_xx")
        self.assertEqual(Q_ux_hash, -1344468598834810301, "Incorrect Q_ux")
        self.assertEqual(Q_uu_hash, -1135300939716028365, "Incorrect Q_uu")

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test5_gains(self):
        """Test gains"""
        # load locals
        gains = self.notebook_locals['gains']

        np.random.seed(7)
        Q_u = np.random.randn(2)
        Q_uu = np.random.randn(2, 2)
        Q_uu = 0.5 * (Q_uu + Q_uu.T)
        Q_ux = np.random.randn(2, 5)
        k, K = gains(Q_uu, Q_u, Q_ux)

        k_hash = hash(tuple(np.ndarray.flatten(self.round(k, 4))))
        K_hash = hash(tuple(np.ndarray.flatten(self.round(K, 4))))

        self.assertEqual(k_hash, -5624267973007621395, "Incorrect k gain")

        self.assertEqual(K_hash, -4694246526328938179, "Incorrect K gain")

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test6_V_terms(self):
        """Test V_terms"""
        # load locals
        V_terms = self.notebook_locals['V_terms']
        gains = self.notebook_locals['gains']

        np.random.seed(7)
        Q_x = np.random.randn(5)
        Q_u = np.random.randn(2)
        Q_uu = np.random.randn(2, 2)
        Q_uu = 0.5 * (Q_uu + Q_uu.T)
        Q_ux = np.random.randn(2, 5)
        Q_xx = np.random.randn(5, 5)
        Q_xx = 0.5 * (Q_xx + Q_xx.T)
        k, K = gains(Q_uu, Q_u, Q_ux)

        V_x, V_xx = V_terms(Q_x, Q_u, Q_xx, Q_ux, Q_uu, K, k)

        V_x_hash = hash(tuple(np.ndarray.flatten(self.round(V_x, 4))))
        V_xx_hash = hash(tuple(np.ndarray.flatten(self.round(V_xx, 4))))

        self.assertEqual(V_x_hash, 2222446538856429062, "Incorrect V_x")

        self.assertEqual(V_xx_hash, 1880932214702358488, "Incorrect V_xx")

    @weight(2)
    @timeout_decorator.timeout(1.)
    def test7_forward_pass(self):
        """Test forward_pass"""
        # load locals
        forward_pass = self.notebook_locals['forward_pass']

        np.random.seed(7)
        N = 5
        n_u = 2
        n_x = 5
        k = np.random.randn(2)
        K = np.random.randn(2, 5)
        u_trj = np.random.randn(N - 1, n_u)
        x_trj = np.random.randn(N, n_x)
        k_trj = np.random.randn(u_trj.shape[0], u_trj.shape[1])
        K_trj = np.random.randn(u_trj.shape[0], u_trj.shape[1], x_trj.shape[1])
        x_trj_new, u_trj_new = forward_pass(x_trj, u_trj, k_trj, K_trj)

        x_trj_new_hash = hash(
            tuple(np.ndarray.flatten(self.round(x_trj_new, 4))))
        u_trj_new_hash = hash(
            tuple(np.ndarray.flatten(self.round(u_trj_new, 4))))

        self.assertEqual(x_trj_new_hash, 8614906876295478816,
                         "Incorrect x_trj_new")

        self.assertEqual(u_trj_new_hash, 5430054884692290987,
                         "Incorrect u_trj_new")

    @weight(2)
    @timeout_decorator.timeout(1.)
    def test8_backward_pass(self):
        """Test backward_pass"""
        # load locals
        backward_pass = self.notebook_locals['backward_pass']

        np.random.seed(7)
        N = 5
        n_u = 2
        n_x = 5
        k = np.random.randn(2)
        K = np.random.randn(2, 5)
        u_trj = np.random.randn(N - 1, n_u)
        x_trj = np.random.randn(N, n_x)
        k_trj, K_trj, expected_cost_redu = backward_pass(x_trj, u_trj, 100.0)

        k_trj_hash = hash(tuple(np.ndarray.flatten(self.round(k_trj, 4))))
        K_trj_hash = hash(tuple(np.ndarray.flatten(self.round(K_trj, 4))))

        self.assertEqual(k_trj_hash, 4400565416540718551, "Incorrect k_trj")

        self.assertEqual(K_trj_hash, 5118135117981053517, "Incorrect K_trj")

    @staticmethod
    def round(n, decimals=0):
        multiplier = 10**decimals
        return np.round(n * multiplier) / multiplier



import asyncio
import os
import sys

from IPython import get_ipython
from IPython.display import display, HTML
from ipywidgets.widgets import FloatSlider
import numpy as np
from warnings import warn

from pydrake.systems.framework import VectorSystem

# Use a global variable here because some calls to IPython will actually case an
# interpreter to be created.  This file needs to be imported BEFORE that
# happens.
running_as_notebook = "COLAB_TESTING" not in os.environ and get_ipython(
) and hasattr(get_ipython(), 'kernel')


def pyplot_is_interactive():
    # import needs to happen after the backend is set.
    import matplotlib.pyplot as plt
    from matplotlib.rcsetup import interactive_bk
    return plt.get_backend() in interactive_bk


# TODO: move this into drake
def ToLatex(M, precision=4):
    M = np.atleast_2d(M)
    md = '\\begin{bmatrix}'
    for row in range(M.shape[0]):
        for col in range(M.shape[1]):
            md += f"{M[row, col]:.{precision}f}"
            if col < M.shape[1] - 1:
                md += " & "
        if row < M.shape[0] - 1:
            md += " \\\\"
    md += '\\end{bmatrix}'
    return md


def AdvanceToAndVisualize(simulator,
                          visualizer,
                          time,
                          time_if_running_headless=0.1):
    """
    Helper to support visualizing a simulation with pyplot visualizer.
    Will simply simulate (with target_realtime_rate = 1) if visualizer.show =
    True, or will record and render an animation if visualizer.show = False. If
    specified, time_if_running_headless will be used instead of time if
    running_as_notebook is False.
    """
    if visualizer._show:
        target_rate = simulator.get_target_realtime_rate()
        simulator.set_target_realtime_rate(1.)
    else:
        print("simulating... ", end=" ")
        visualizer.start_recording()
    if time_if_running_headless and not running_as_notebook:
        time = time_if_running_headless
    simulator.AdvanceTo(time)
    if visualizer._show:
        simulator.set_target_realtime_rate(target_rate)
    else:
        print("done.\ngenerating animation...")
        ani = visualizer.get_recording_as_animation()
        display(HTML(ani.to_jshtml()))


def SetupMatplotlibBackend(wishlist=["notebook"]):
    """
    Helper to support multiple workflows:
        1) nominal -- running locally w/ jupyter notebook
        2) unit tests (no ipython, backend is template)
        3) binder -- does have notebook backend
        4) colab -- claims to have notebook, but it doesn't work
    Puts the matplotlib backend into notebook mode, if possible,
    otherwise falls back to inline mode.
    Returns True iff the final backend is interactive.
    """
    # To find available backends, one can access the lists:
    # matplotlib.rcsetup.interactive_bk
    # matplotlib.rcsetup.non_interactive_bk
    # matplotlib.rcsetup.all_backends
    if running_as_notebook:
        ipython = get_ipython()
        # Short-circuit for google colab.
        if 'google.colab' in sys.modules:
            ipython.run_line_magic("matplotlib", "inline")
            return False
        # TODO: Find a way to detect vscode, and use inline instead of notebook
        for backend in wishlist:
            try:
                ipython.run_line_magic("matplotlib", backend)
                return pyplot_is_interactive()
            except KeyError:
                continue
        ipython.run_line_magic("matplotlib", "inline")
    return False


# Deprecate everything below this line:


def setup_matplotlib_backend(**kwargs):
    SetupMatplotlibBackend(**kwargs)


# Inspired by https://github.com/Kirill888/jupyter-ui-poll but there *must* be a
# better way. Ideas:
#  - provide init() and cleanup() methods that could be called to pull out and
#    replace the irrelevant events just once (instead of on every timestep).
#    E.g. init(), simulator.AdvanceTo(10), cleanup().
#  - whitelist widget events and only process them (instead of black-listing the
#    execute_request).
#  - do I actually need asyncio?  can I just write the events back?
#
# But I'll wait to see how much we use this before spending too much time on it.
# BTW,
# https://ipywidgets.readthedocs.io/en/latest/examples/Widget%20Asynchronous.html  # noqa
# describes the problem but does not offer a solution.
def update_widgets(num_ui_events_to_process=1):
    shell = get_ipython()
    # Ok to do nothing if running from console
    if shell is None:
        return
    kernel = shell.kernel
    events = []
    kernel.shell_handlers['execute_request'] = lambda *e: events.append(e)
    current_parent = (kernel._parent_ident, kernel._parent_header)

    for _ in range(num_ui_events_to_process):
        # ensure stdout still happens in the same cell
        kernel.set_parent(*current_parent)
        kernel.do_one_iteration()
        kernel.set_parent(*current_parent)

    kernel.shell_handlers['execute_request'] = kernel.execute_request

    def _replay_events(shell, events):
        kernel = shell.kernel
        sys.stdout.flush()
        sys.stderr.flush()
        for stream, ident, parent in events:
            kernel.set_parent(ident, parent)
            if kernel._aborting:
                kernel._send_abort_reply(stream, parent, ident)
            else:
                kernel.execute_request(stream, ident, parent)

    loop = asyncio.get_event_loop()
    if loop.is_running():
        loop.call_soon(lambda: _replay_events(shell, events))
    else:
        warn('Automatic execution of scheduled cells only works with '
             'asyncio-based ipython')


# TODO(russt): generalize this to e.g. WidgetSystem (should work for any widget,
# or list of widgets).
class SliderSystem(VectorSystem):

    def __init__(self, min, max, value=0, description=""):
        # 0 inputs, 1 output.
        VectorSystem.__init__(self, 0, 1)
        self.slider = FloatSlider(value=value,
                                  description=description,
                                  min=min,
                                  max=max,
                                  continuous_update=True)

        if get_ipython() is not None:
            display(self.slider)

    def DoCalcVectorOutput(self, context, unused, unused2, output):
        update_widgets()
        output[:] = self.slider.value



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestLearningDynamics(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(5)
    @timeout_decorator.timeout(10.)
    def test_final_loss(self):
        """Test dynamics model training"""
        # note: all prints here go to the output item in the json file

        # the loss must be close to the correct value
        final_loss = self.notebook_locals['final_loss']
        self.assertLessEqual(
            final_loss,
            0.38,
            msg='Final loss is less than upper bound for the correct value')
        self.assertGreaterEqual(
            final_loss,
            0.35,
            msg='Final loss is greater than lower bound for the correct value')



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestHopfield(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(6)
    @timeout_decorator.timeout(5.)
    def test_Hopfield_Network(self):
        """Test Hopfield Network"""
        # note: all prints here go to the output item in the json file

        dynamics = self.notebook_locals['dynamics']
        calc_A = self.notebook_locals['calculate_A']

        A = np.array([[1, 0], [0, 1], [-1, -1]])
        beta = 5

        x = np.array([0, 0])
        x_target = np.array([0, 0])
        x_nxt = dynamics(x, A, beta)

        self.assertTrue(
            all(abs(i - j) <= 1e-4 for i, j in zip(x_nxt, x_target)),
            msg='For x = [0, 0], x_dot should be {}'.format(x_target),
        )

        x = np.array([1, 0])
        x_target = np.array([-0.00678274, 0.00664746])
        x_nxt = dynamics(x, A, beta)

        self.assertTrue(
            all(abs(i - j) <= 1e-4 for i, j in zip(x_nxt, x_target)),
            msg='For x = [1, 0], x_dot should be {}'.format(x_target),
        )

        x = np.array([1, 1])
        x_target = np.array([-0.50000023, -0.50000023])
        x_nxt = dynamics(x, A, beta)

        self.assertTrue(
            all(abs(i - j) <= 1e-4 for i, j in zip(x_nxt, x_target)),
            msg='Answer is incorrect for x = [1, 1].',
        )

        A = []
        for item in self.notebook_locals['training_data']:
            A += [item.reshape((-1))]
        A_target = np.array(A)
        A_calc = calc_A(self.notebook_locals['training_data'])

        self.assertTrue(
            all(
                abs(i - j) <= 1e-4
                for i, j in zip(A_target.shape, A_calc.shape)),
            msg='Expecting the shape of A to be {}, but got {}.'.format(
                A_target.shape, A_calc.shape),
        )

        A_calc = calc_A(
            [np.zeros((28, 28)).reshape((-1)),
             np.ones((28, 28)).reshape((-1))])
        A_target1 = np.array(
            [np.zeros((28, 28)).reshape((-1)),
             np.ones((28, 28)).reshape((-1))])
        A_target2 = np.array(
            [np.ones((28, 28)).reshape((-1)),
             np.zeros((28, 28)).reshape((-1))])

        diff_a = (abs(A_target1 - A_calc)).max() > 1e-4
        diff_b = (abs(A_target2 - A_calc)).max() > 1e-4
        self.assertTrue(diff_a or diff_b,
                        msg='calculate_A did not produce expected output.')



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestFittedDoubleIntegrator(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(2)
    @timeout_decorator.timeout(20.)
    def test_final_losses(self):
        """Test model training"""
        # note: all prints here go to the output item in the json file

        # the loss must be close to the correct value
        final_loss_min_time = self.notebook_locals['final_loss_min_time']
        self.assertLessEqual(
            final_loss_min_time,
            0.0003,
            msg='Min time loss is < than upper bound for the correct value')
        self.assertGreaterEqual(
            final_loss_min_time,
            0.0001,
            msg='Min time loss is > than lower bound for the correct value')

        final_loss_quadratic = self.notebook_locals['final_loss_quadratic']
        self.assertLessEqual(
            final_loss_quadratic,
            0.009,
            msg='Quadratic loss is < than upper bound for the correct value')
        self.assertGreaterEqual(
            final_loss_quadratic,
            0.002,
            msg='Quadratic loss is > than lower bound for the correct value')

    @weight(2)
    @timeout_decorator.timeout(20.)
    def test_policy(self):
        """Test closed form policy"""
        # note: all prints here go to the output item in the json file
        final_state_min_time = self.notebook_locals['final_state_min_time']
        self.assertLessEqual(
            final_state_min_time[0],
            0.6,
            msg='Minimum time policy final position is too large')
        self.assertGreaterEqual(
            final_state_min_time[0],
            -0.6,
            msg='Minimum time policy final position is too small')
        self.assertLessEqual(
            final_state_min_time[1],
            0.5,
            msg='Minimum time policy final velocity is too large')
        self.assertGreaterEqual(
            final_state_min_time[1],
            -0.5,
            msg='Minimum time policy final velocity is too small')

        final_state_quadratic = self.notebook_locals['final_state_quadratic']
        self.assertLessEqual(final_state_quadratic[0],
                             0.6,
                             msg='Quadratic policy final position is too large')
        self.assertGreaterEqual(
            final_state_quadratic[0],
            -0.6,
            msg='Quadratic policy final position is too small')
        self.assertLessEqual(final_state_quadratic[1],
                             0.5,
                             msg='Quadratic policy final velocity is too large')
        self.assertGreaterEqual(
            final_state_quadratic[1],
            -0.5,
            msg='Quadratic policy final velocity is too small')



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestLinearSysid(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(3)
    @timeout_decorator.timeout(20.)
    def test_L2_regression(self):
        """Test L2 regression"""
        Ahat = self.notebook_locals['Ahat_2norm']
        Bhat = self.notebook_locals['Bhat_2norm']
        X = self.notebook_locals['X']
        U = self.notebook_locals['U']
        res = Ahat @ X[:, :-1] + Bhat @ U[:, :-1] - X[:, 1:]
        self.assertLessEqual(
            np.sum(res**2), 6e-7,
            'The residual for 2-norm minimization is too high!')

    @weight(3)
    @timeout_decorator.timeout(20.)
    def test_Linf_regression(self):
        """Test L infinite regression"""
        Ahat = self.notebook_locals['Ahat_infnorm']
        Bhat = self.notebook_locals['Bhat_infnorm']
        X = self.notebook_locals['X']
        U = self.notebook_locals['U']
        res = Ahat @ X[:, :-1] + Bhat @ U[:, :-1] - X[:, 1:]
        residual = np.sum(np.max(np.abs(res), axis=0))
        self.assertLessEqual(
            residual, 0.02,
            'The residual for infinity-norm minimization is too high!')

        obj_infnorm = self.notebook_locals['obj_infnorm']
        self.assertLessEqual(
            np.abs(obj_infnorm - residual), 0.1,
            'Objective from optimizer and empirical evaluations are different!')

    @weight(3)
    @timeout_decorator.timeout(20.)
    def test_L1_regression(self):
        """Test L1 regression"""
        Ahat = self.notebook_locals['Ahat_1norm']
        Bhat = self.notebook_locals['Bhat_1norm']
        X = self.notebook_locals['X']
        U = self.notebook_locals['U']
        res = Ahat @ X[:, :-1] + Bhat @ U[:, :-1] - X[:, 1:]
        residual = np.sum(np.abs(res))
        self.assertLessEqual(
            residual, 0.04, 'The residual for 1-norm minimization is too high!')

        obj_1norm = self.notebook_locals['obj_1norm']
        self.assertLessEqual(
            np.abs(obj_1norm - residual), 0.1,
            'Objective from optimizer and empirical evaluations are different!')



import unittest
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
import numpy as np


class TestShootingVsTranscription(unittest.TestCase):

    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test_lost_bits_shooting(self):
        """Test number of bits lost with direct shooting"""
        # get list of lost bits
        lost_bits = self.notebook_locals['lost_bits_shooting']

        # check list length and types
        self.assertEqual(
            len(lost_bits), 45,
            'The list lost_bits_shooting must have length equal' + 'to 45.')
        bits_type = [isinstance(bits, (float, int)) for bits in lost_bits]
        self.assertTrue(
            all(bits_type), 'Some entries in lost_bits_shooting are neither'
            + 'floats nor ints.')

        # check values
        lost_bits_shooting = [
            0.0, 0.9543364613711909, 1.8123366596726525, 2.845325160323255,
            3.997099761655346, 5.181146076477944, 6.367665432089994,
            7.549745045378025, 8.727297345452408, 9.901650046513046,
            11.074027866040504, 12.245271423678894, 13.415892162590488,
            14.586181089218373, 15.75629682473095, 16.92632331212132,
            18.096304057604645, 19.266261289075352, 20.436206266052793,
            21.606144677268965, 22.776079410190473, 23.94601195116546,
            25.115943085840197, 26.285873247157944, 27.455802687733357,
            28.625731565680532, 29.795659987345097, 30.965588028826133,
            32.13551574866848, 33.30544319406715, 34.47537039717343,
            35.64529738697145, 36.81522421237407, 37.98515087701531,
            39.15507737686514, 40.32500371458269, 41.49492968101353,
            42.6648551872069, 43.83478032928076, 45.00470555237002,
            46.174631383871656, 47.34456378440541, 48.52985492074265,
            49.72208993747556, 50.8608777068896
        ]
        for i, bits in enumerate(lost_bits):
            self.assertAlmostEqual(bits,
                                   lost_bits_shooting[i],
                                   places=1,
                                   msg='lost_bits_shooting are incorrect.')

    @weight(4)
    @timeout_decorator.timeout(1.)
    def test_Mb_diagonal_block(self):
        """Test construction of diagonal block of Mb"""
        # get function to construct the block
        Mb_diagonal_block = self.notebook_locals['Mb_diagonal_block']

        # test function directly on the problem data
        A = np.array([[1, .5], [.5, 1]])
        B = np.array([[0], [.5]])
        Q = np.array([[1, 0], [0, 0]])
        R = np.array([[1]])

        # run sudent function
        Mb_block = Mb_diagonal_block(A, B, Q, R).toarray()

        # correct solution for the fake data
        Mb_block_target = np.array([
            [-1, 0, 2, 0, 0, 1, .5, 0, 0],
            [0, -1, 0, 0, 0, .5, 1, 0, 0],
            [0, 0, 0, 0, 2, 0, .5, 0, 0],
            [0, 0, 1, .5, 0, 0, 0, -1, 0],
            [0, 0, .5, 1, .5, 0, 0, 0, -1],
        ])
        self.assertTrue(
            Mb_block.shape == Mb_block_target.shape,
            'The matrix we got from Mb_diagonal_block has wrong dimensions.')
        np.testing.assert_array_almost_equal(
            Mb_block,
            Mb_block_target,
            err_msg='The function Mb_diagonal_block is incorrect on the '
            + f'problem data A = {A}, B = {B}, Q = {Q}, and R = {R}.')

        # test function directly on synthtic data
        A = np.arange(9).reshape(3, 3)
        B = np.arange(6).reshape(3, 2)
        Q = np.ones((3, 3)) * 2
        R = np.ones((2, 2)) * 3

        # run sudent function
        Mb_block = Mb_diagonal_block(A, B, Q, R).toarray()

        # correct solution for the fake data
        Mb_block_target = np.array([
            [-1, 0, 0, 4, 4, 4, 0, 0, 0, 3, 6, 0, 0, 0],
            [0, -1, 0, 4, 4, 4, 0, 0, 1, 4, 7, 0, 0, 0],
            [0, 0, -1, 4, 4, 4, 0, 0, 2, 5, 8, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 6, 6, 0, 2, 4, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 6, 6, 1, 3, 5, 0, 0, 0],
            [0, 0, 0, 0, 1, 2, 0, 1, 0, 0, 0, -1, 0, 0],
            [0, 0, 0, 3, 4, 5, 2, 3, 0, 0, 0, 0, -1, 0],
            [0, 0, 0, 6, 7, 8, 4, 5, 0, 0, 0, 0, 0, -1],
        ])
        self.assertTrue(
            Mb_block.shape == Mb_block_target.shape,
            'The matrix we got from Mb_diagonal_block has wrong dimensions.')
        np.testing.assert_array_almost_equal(
            Mb_block,
            Mb_block_target,
            err_msg='The function Mb_diagonal_block is incorrect on the '
            + f'problem data A = {A}, B = {B}, Q = {Q}, and R = {R}.')

    @weight(4)
    @timeout_decorator.timeout(1.)
    def test_J_star_N_transcription(self):
        """Test value function direct transcription"""
        # get list of costs to go
        J_star_N = self.notebook_locals['J_star_N_transcription']

        # check list length and types
        self.assertEqual(
            len(J_star_N), 100,
            'The list J_star_N_transcription must have length'
            + 'equal to 100.')
        J_star_type = [isinstance(J, (float, int)) for J in J_star_N]
        self.assertTrue(
            all(J_star_type),
            'Some entries in J_star_N_transcription are neither'
            + 'floats nor ints.')

        # check values
        self._test_J_star(J_star_N, 'J_star_N_transcription is incorrect.')

    @weight(3)
    @timeout_decorator.timeout(1.)
    def test_lost_bits_transcription(self):
        """Test number of bits lost with direct transcription"""
        # get list of lost bits
        lost_bits = self.notebook_locals['lost_bits_transcription']

        # check list length and types
        self.assertEqual(
            len(lost_bits), 100,
            'The list lost_bits_transcription must have length'
            + 'equl to 100.')
        bits_type = [isinstance(bits, (float, int)) for bits in lost_bits]
        self.assertTrue(
            all(bits_type),
            'Some entries in lost_bits_transcription are neither'
            + 'floats nor ints.')

        # check values
        lost_bits_transcription_30 = [
            6.659823518580827, 7.566530101329372, 7.989788581807347,
            8.126275184924792, 8.156165618892503, 8.158803371775676,
            8.1571939291375, 8.155956959505353, 8.155384990230832,
            8.155170750799739, 8.155101341468276, 8.155081672171026,
            8.155076924745707, 8.155076049611791, 8.15507599036841,
            8.155076036096093, 8.155076066576788, 8.155076079192476,
            8.155076083345309, 8.1550760844654, 8.155076084693125,
            8.155076084711247, 8.155076084698281, 8.155076084688659,
            8.15507608468425, 8.15507608468261, 8.155076084682092,
            8.15507608468194, 8.155076084681896, 8.155076084681886
        ]
        for i, bits in enumerate(lost_bits):
            self.assertAlmostEqual(bits,
                                   lost_bits_transcription_30[min(i, 29)],
                                   msg='lost_bits_transcription are incorrect.')

    @weight(4)
    @timeout_decorator.timeout(1.)
    def test_J_star_N_riccati(self):
        """Test value function Riccati recursion"""
        # get list of costs to go
        J_star_N = self.notebook_locals['J_star_N_riccati']

        # check list length and types
        self.assertEqual(
            len(J_star_N), 1000,
            'The list J_star_N_riccati must have length equal' + 'to 1000.')
        J_star_type = [isinstance(J, (float, int)) for J in J_star_N]
        self.assertTrue(
            all(J_star_type),
            'Some entries in J_star_N_riccati are neither floats' + 'nor ints.')

        # check values
        self._test_J_star(J_star_N, 'J_star_N_riccati is incorrect.')

    def _test_J_star(self, J_star_N, msg):

        # check values
        J_star_N_30 = [
            10.75, 18.506849315068493, 24.221925133689837, 26.538390153365242,
            27.14309163681062, 27.243582216689365, 27.242000076428884,
            27.231499085690622, 27.225612316892573, 27.223217461345843,
            27.222393413120585, 27.222143703386408, 27.222076887248043,
            27.222061591713505, 27.22205892953035, 27.22205878066236,
            27.222058926005232, 27.222059017397356, 27.22205905444766,
            27.222059066651646, 27.222059070055877, 27.22205907083573,
            27.222059070957876, 27.222059070953733, 27.22205907093976,
            27.22205907093211, 27.22205907092905, 27.222059070928,
            27.222059070927678, 27.222059070927585
        ]
        for i, J in enumerate(J_star_N):
            self.assertAlmostEqual(J, J_star_N_30[min(i, 29)], msg=msg)



from pydrake.autodiffutils import AutoDiffXd
from pydrake.multibody.tree import MultibodyForces_
from pydrake.multibody.plant import MultibodyPlant_
from pydrake.symbolic import Expression
import numpy as np


def ManipulatorDynamics(plant, q, v=None):
    context = plant.CreateDefaultContext()
    plant.SetPositions(context, q)
    if v is not None:
        plant.SetVelocities(context, v)
    M = plant.CalcMassMatrixViaInverseDynamics(context)
    Cv = plant.CalcBiasTerm(context)
    tauG = plant.CalcGravityGeneralizedForces(context)
    B = plant.MakeActuationMatrix()
    forces = MultibodyForces_(plant)
    plant.CalcForceElementsContribution(context, forces)
    # TODO(russt): add in contact forces to tauExt.
    tauExt = forces.generalized_forces()

    return (M, Cv, tauG, B, tauExt)



from matplotlib.widgets import Slider
from matplotlib.animation import HTMLWriter

from pydrake.systems.framework import VectorSystem


class SliderSystem(VectorSystem):

    def __init__(self, ax, title, min, max):
        # 0 inputs, 1 output.
        VectorSystem.__init__(self, 0, 1)
        self.value = 0
        self.slider = Slider(ax, title, min, max, valinit=self.value)
        self.slider.on_changed(self.update)

    def update(self, val):
        self.value = val

    def DoCalcVectorOutput(self, context, unused, unused2, output):
        output[:] = self.value


def AdvanceToAndSaveAnimation(simulator, visualizer, time, filename):
    visualizer.start_recording()
    simulator.AdvanceTo(time)
    ani = visualizer.get_recording_as_animation()
    # Note: Wanted to use embed_frames=True, but it did not render
    # the image properly for me.
    writer = HTMLWriter()
    writer.frame_format = 'svg'
    ani.save(filename, writer=writer)
